from __future__ import annotations
from typing import List, Union
from .section import Section
from .zeros import Zeros

import logging

LOG = logging.getLogger(__name__)


class Packet(List[Section]):
    """Represents a packet as a list of sections, it extends on list itself so all methods for lists can be used"""

    def to_int(self) -> int:
        return int.from_bytes(self.to_bytes(), "big")

    def to_bytes(self) -> bytes:
        result = bytearray()
        for section in self:
            result.extend(section.to_bytes())
        return bytes(result)

    def byte_size(self) -> int:
        return sum(section.byte_size for section in self)

    def decode(
        self,
        bytes_value: Union[bytes, bytearray],
        trail_or_lead: Zeros = Zeros.TRAILING,
    ) -> Packet:
        value = bytes(bytes_value)
        try:
            value = Zeros.insert(trail_or_lead, bytes_value, self.byte_size())
        except OverflowError as err:
            if next(filter(lambda section: section.byte_size == 0, self), None):
                LOG.debug(
                    "overflow error while parsing, ignoring the error since there might be expanding sections"
                )
            else:
                raise err
        for section in self:
            if len(value) == 0:
                return self
            value = section.partial_decode(value)
        return self

    def __str__(self) -> str:
        # each byte has two chars in hex representation
        return "0x{0:0{1}X}".format(self.to_int(), self.byte_size() * 2)



from .section import Section
from .zeros import Zeros


class ExpandingSection(Section):
    """Represents a section that expands while decoding or setting bytes"""

    def __init__(self):
        self.value = 0
        self.byte_size = 0
        self.trail_or_lead = Zeros.LEADING

    def set_bytes(self, bytes_value: bytes) -> Section:
        """sets the value and the byte size according to values max size"""
        self.byte_size = len(bytes_value)
        super().set_bytes(bytes_value)
        return self

    def partial_decode(self, bytes_value: bytes) -> bytes:
        """decodes all of the given bytes and set it as value while updating its own byte size. Returns empty list"""
        self.set_bytes(bytes_value)
        return bytes()



# __init__.py
from .packet import Packet as Packet
from .section import Section as Section
from .expanding_section import ExpandingSection as ExpandingSection
from .dynamic_section import DynamicSection as DynamicSection
from .zeros import Zeros as Zeros
from .errors import byte_size_error as byte_size_error
from .errors import value_overflow_error as value_overflow_error



from __future__ import annotations
from typing import Any
from math import ceil
from .zeros import Zeros
from .errors import byte_size_error, value_overflow_error


class Section:
    """Represents a section of a packet"""

    __slots__ = ["value", "byte_size", "trail_or_lead"]

    @staticmethod
    def Template(byte_size: int, trail_or_lead: Zeros = Zeros.LEADING) -> Section:
        return Section(0, byte_size, trail_or_lead)

    @staticmethod
    def _minimum_byte_size(value: int) -> int:
        size = ceil(value.bit_length() / 8)
        if size == 0:
            return 1
        return size

    def __init__(
        self, value: int = 0, byte_size: int = 0, trail_or_lead: Zeros = Zeros.LEADING
    ) -> None:
        value_size = self._minimum_byte_size(value)
        if byte_size == 0:
            byte_size = value_size
        if byte_size < 0:
            raise byte_size_error
        if value_size > byte_size:
            raise value_overflow_error
        self.value = value
        self.byte_size = byte_size
        self.trail_or_lead = trail_or_lead

    def to_bytes(self) -> bytes:
        """returns value as bytes, if needed inserts zeros according the initialized leading or trailing zeros config"""
        return Zeros.insert(
            self.trail_or_lead,
            self.value.to_bytes(self.byte_size, byteorder="big"),
            self.byte_size,
        )

    def set_bytes(self, bytes_value: bytes) -> Section:
        """sets only the value for the section from given bytes"""
        value = Zeros.insert(self.trail_or_lead, bytes_value, self.byte_size)
        new_value = int.from_bytes(value, "big")
        new_value_size = self._minimum_byte_size(new_value)
        if (
            not (self.byte_size == 0 and new_value == 0)
            and new_value_size > self.byte_size
        ):
            raise value_overflow_error
        self.value = new_value
        return self

    def partial_decode(self, bytes_value: bytes) -> bytes:
        """decodes only the relevant part of a given byte array and returns the remaining bytes, starts from leftmost byte"""
        max_len = len(bytes_value)
        if self.byte_size < max_len:
            max_len = self.byte_size
        self.set_bytes(bytes_value[0:max_len])
        return bytes_value[max_len:]

    def __str__(self) -> str:
        # each byte has two chars in hex representation
        return "0x{0:0{1}X}".format(self.value, self.byte_size * 2)

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, other: Any):
        return (
            isinstance(other, self.__class__)
            and self.value == other.value
            and self.byte_size == other.byte_size
        )

    def __hash__(self) -> int:
        return hash((self.value, self.byte_size))



from __future__ import annotations
from typing import Union
from enum import Enum
from .errors import value_overflow_error


class Zeros(Enum):
    TRAILING = 1
    LEADING = 2

    @staticmethod
    def insert(zeros: Zeros, data: Union[bytes, bytearray], byte_size: int) -> bytes:
        value = bytes(data)
        if len(data) > byte_size:
            raise value_overflow_error
        diff = byte_size - len(value)
        if zeros == Zeros.TRAILING:
            value = value + bytes([0 for _ in range(diff)])
        else:
            value = bytes([0 for _ in range(diff)]) + value
        return value



from .section import Section
from .zeros import Zeros
from typing import Any
from .errors import byte_size_error, value_overflow_error


class DynamicSection(Section):
    """Represents a section that changes size according to the first byte of decoding value
    the first byte is called length byte because it determines the length of bytes that needs
    to be read and set as value. The length byte's size can be set with length_byte_size
    set_bytes for this class only sets the value without the size byte so input should not include size byte
    """

    __slots__ = ["length_byte_size"]

    def __init__(
        self, length_byte_size: int = 1, trail_or_lead: Zeros = Zeros.TRAILING
    ) -> None:
        if length_byte_size <= 0:
            raise byte_size_error
        super().__init__()
        self.value = 0
        self.byte_size = 0
        self.trail_or_lead = trail_or_lead
        self.length_byte_size = length_byte_size

    def set_bytes(self, bytes_value: bytes) -> Section:
        """sets the value and the byte size according to value size for the section from given bytes, input shouldn't include length bytes
        if length of value doesn't fit into the length byte size which is set during initialization (length_byte_size), it raises overflow error
        """
        self.byte_size = len(bytes_value)
        if self._minimum_byte_size(self.byte_size) > self.length_byte_size:
            raise value_overflow_error
        super().set_bytes(bytes_value)
        return self

    def to_bytes(self) -> bytes:

        """returns length of value as bytes and value as bytes appended, if needed inserts zeros according the initialized leading or trailing zeros config
        length bytes are always the first byte independent of the trailing or leading zeros config
        """

        return (
            Section(value=self.byte_size, byte_size=self.length_byte_size).to_bytes()
            + super().to_bytes()
        )

    def partial_decode(self, bytes_value: bytes) -> bytes:

        """decodes given bytes into byte size and value, start from left most
        first n bytes determines the byte size, n being the length byte size which is set during initialization (length_byte_size)
        the value in the first n bytes determines the byte size which determines how many bytes to be read and saved as section value
        the remaining bytes after byte size is read are returned
        """

        length_section = Section.Template(self.length_byte_size)
        bytes_value = length_section.partial_decode(bytes_value)
        self.byte_size = length_section.value
        max_len = len(bytes_value)
        if self.byte_size < max_len:
            max_len = self.byte_size
        partial_value = bytes_value[0:max_len]
        super().set_bytes(partial_value)
        return bytes_value[max_len:]

    def __str__(self) -> str:
        # each byte has two chars in hex representation
        return (
            "0x{0:0{1}X}".format(self.byte_size, self.length_byte_size * 2)
            + "-"
            + "0x{0:0{1}X}".format(self.value, self.byte_size * 2)
        )

    def __eq__(self, other: Any):
        return (
            isinstance(other, self.__class__)
            and self.value == other.value
            and self.byte_size == other.byte_size
            and self.length_byte_size == other.length_byte_size
        )

    def __hash__(self) -> int:
        return hash((self.value, self.byte_size, self.length_byte_size))
